/*
 * Asterisk -- An open source telephony toolkit.
 *
 * Copyright (C) 1999 - 2005, Digium, Inc.
 *
 * Mark Spencer <markster@digium.com>
 *
 * See http://www.asterisk.org for more information about
 * the Asterisk project. Please do not directly contact
 * any of the maintainers of this project for assistance;
 * the project provides a web site, mailing lists and IRC
 * channels for your use.
 *
 * This program is free software, distributed under the terms of
 * the GNU General Public License Version 2. See the LICENSE file
 * at the top of the source tree.
 */

/*! \file
 *
 * \brief Trivial application to read an extension into a variable
 *
 * \author David Chappell <David.Chappell@trincoll.edu>
 * 
 * \ingroup applications
 */
 
#include "asterisk.h"

ASTERISK_FILE_VERSION(__FILE__, "$Revision: 42298 $")

#include <stdlib.h>
#include <stdio.h>
#include <string.h>

#include "asterisk/lock.h"
#include "asterisk/file.h"
#include "asterisk/logger.h"
#include "asterisk/channel.h"
#include "asterisk/pbx.h"
#include "asterisk/app.h"
#include "asterisk/module.h"
#include "asterisk/translate.h"
#include "asterisk/options.h"
#include "asterisk/utils.h"
#include "asterisk/indications.h"

enum {
	OPT_SKIP = (1 << 0),
	OPT_INDICATION = (1 << 1),
	OPT_NOANSWER = (1 << 2),
} readexten_option_flags;

AST_APP_OPTIONS(readexten_app_options, {
	AST_APP_OPTION('s', OPT_SKIP),
	AST_APP_OPTION('i', OPT_INDICATION),
	AST_APP_OPTION('n', OPT_NOANSWER),
});

static char *app = "ReadExten";

static char *synopsis = "Read an extension into a variable";

static char *descrip = 
"  ReadExten(variable[|filename][|context][|option][|timeout])\n\n"
"Reads a #-terminated string of digits a certain number of times from the\n"
"user into the given variable.\n"
"  filename   -- file to play before reading digits or tone with option i\n"
"  context    -- context in which to match extensions\n"
"  option     -- options are 's' , 'i', 'n'\n"
"                's' to return immediately if the line is not up,\n"
"                'i' to play  filename as an indication tone from your indications.conf\n"
"                'n' to read digits even if the line is not up.\n"
"  timeout    -- An integer number of seconds to wait for a digit response. If greater\n"
"                than 0, that value will override the default timeout.\n\n"
"ReadExten should disconnect if the function fails or errors out.\n";


#define ast_next_data(instr,ptr,delim) if((ptr=strchr(instr,delim))) { *(ptr) = '\0' ; ptr++;}

static int readexten_exec(struct ast_channel *chan, void *data)
{
	int res = 0;
	struct ast_module_user *u;
	char exten[256] = "";
	int maxdigits = 255;
	int timeout = 0, digit_timeout = 0, x = 0;
	char *argcopy = NULL;
	struct tone_zone_sound *ts;
	struct ast_flags flags = {0};

	 AST_DECLARE_APP_ARGS(arglist,
		AST_APP_ARG(variable);
		AST_APP_ARG(filename);
		AST_APP_ARG(context);
		AST_APP_ARG(options);
		AST_APP_ARG(timeout);
	);
	
	if (ast_strlen_zero(data)) {
		ast_log(LOG_WARNING, "ReadExten requires at least one argument\n");
		return -1;
	}

	u = ast_module_user_add(chan);
	
	argcopy = ast_strdupa(data);

	AST_STANDARD_APP_ARGS(arglist, argcopy);

	if (ast_strlen_zero(arglist.variable)) {
		ast_log(LOG_WARNING, "Invalid! Usage: ReadExten(variable[|filename][|context][|options][|timeout])\n\n");
		ast_module_user_remove(u);
		return -1;
	}

	if (ast_strlen_zero(arglist.filename))
		arglist.filename = NULL;

	if (ast_strlen_zero(arglist.context))
		arglist.context = chan->context;

	if (!ast_strlen_zero(arglist.options))
		ast_app_parse_options(readexten_app_options, &flags, NULL, arglist.options);

	if (!ast_strlen_zero(arglist.timeout)) {
		timeout = atoi(arglist.timeout);
		if(timeout > 0)
			timeout *= 1000;
	}

	if (timeout <= 0)
		timeout = chan->pbx ? chan->pbx->rtimeout * 1000 : 10000;

	if (digit_timeout <= 0)
		digit_timeout = chan->pbx ? chan->pbx->dtimeout * 1000 : 5000;

	ts = NULL;
	if (ast_test_flag(&flags,OPT_INDICATION) && !ast_strlen_zero(arglist.filename))
			ts = ast_get_indication_tone(chan->zone,arglist.filename);

	if (chan->_state != AST_STATE_UP) {
		if (ast_test_flag(&flags,OPT_SKIP)) {
			/* At the user's option, skip if the line is not up */
			pbx_builtin_setvar_helper(chan, arglist.variable, "");
			ast_module_user_remove(u);
			return 0;
		} else if (!ast_test_flag(&flags,OPT_NOANSWER)) {
			/* Otherwise answer unless we're supposed to read while on-hook */
			res = ast_answer(chan);
		}
	}

	if (res >= 0) {
		ast_playtones_stop(chan);
		ast_stopstream(chan);

		if (ts && ts->data[0])
			res = ast_playtones_start(chan, 0, ts->data, 0);
		else if(arglist.filename)
			res = ast_streamfile(chan, arglist.filename, chan->language);

		for (x = 0; x < maxdigits; ) {
			ast_verbose(VERBOSE_PREFIX_3 "extension so far: '%s', timeout: %d\n", exten, timeout);
			res = ast_waitfordigit(chan, timeout);

			ast_playtones_stop(chan);
			ast_stopstream(chan);
			timeout = digit_timeout;

			if (res < 1) {		/* timeout expired or hangup */
				exten[x]='\0';
				break;
			}

			if(res == '#') {
				break;
			}

			exten[x++] = res;
			exten[x] = '\0';
			if(!ast_matchmore_extension(chan,arglist.context,exten,1,chan->cid.cid_num)) {
				break;
			}
		}
	}

	if (res < 0) {		/* hangup */
		if (option_verbose > 2)
			ast_verbose(VERBOSE_PREFIX_3 "User disconnected\n");
		pbx_builtin_setvar_helper(chan, arglist.variable, "");
	} else {
		if (ast_strlen_zero(exten)) {	/* empty answer */
			if (res == 0) {				/* entered nothing, not even a #, */
				if (ast_exists_extension(chan, chan->context, "t", 1, chan->cid.cid_num)) {
					if (option_verbose > 2)
						ast_verbose(VERBOSE_PREFIX_3 "Timeout on %s\n", chan->name);
				    ast_copy_string(chan->exten, "t", sizeof(chan->exten));
   					chan->priority = 0;
				} else {
					ast_log(LOG_WARNING, "Timeout, but no rule 't' in context '%s'\n", chan->context);
				}
			} else {
				ast_verbose(VERBOSE_PREFIX_3 "User entered nothing.\n");
				pbx_builtin_setvar_helper(chan, arglist.variable, "");
			}
		} else if(ast_exists_extension(chan, arglist.context, exten, 1, chan->cid.cid_num)) {
			if (option_verbose > 2)
				ast_verbose(VERBOSE_PREFIX_3 "User entered valid extension '%s'\n", exten);
			pbx_builtin_setvar_helper(chan, arglist.variable, exten);
		} else {
			if (ast_exists_extension(chan, chan->context, "i", 1, chan->cid.cid_num)) {
				if (option_verbose > 2)
					ast_verbose(VERBOSE_PREFIX_3 "User dialed invalid extension '%s' in context '%s' on %s\n", exten, arglist.context, chan->name);
				pbx_builtin_setvar_helper(chan, "INVALID_EXTEN", chan->exten);
			    ast_copy_string(chan->exten, "i", sizeof(chan->exten));
   				chan->priority = 0;
			} else {
				ast_log(LOG_WARNING, "User dialed invalid extension '%s' in context '%s', but no invalid handler in context '%s' on %s\n",
					exten, arglist.context, chan->context, chan->name);
			}
		}
	res = 0;
	}

	ast_module_user_remove(u);
	return res;
}

static int unload_module(void)
{
	int res;

	res = ast_unregister_application(app);
	
	ast_module_user_hangup_all();

	return res;	
}

static int load_module(void)
{
	return ast_register_application(app, readexten_exec, synopsis, descrip);
}

AST_MODULE_INFO_STANDARD(ASTERISK_GPL_KEY, "Let the user enter an extension and store in variable");
